#include <algorithm>
#include <cmath>
#include <cstdlib>

#include "common.h"

namespace inferllm {
namespace opt {
#if 0
size_t vlmax[4][4];
size_t vlmul[4][4];

void init() {
    int a;
    asm volatile("csrr %0, vlenb" : "=r"(a));
    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            vlmax[i][j] = a / (1 << i) * (1 << j);
            vlmul[i][j] = j;
        }
    }
}
#endif
void dumpCSR() {
#define CSR(a)                                                                      \
    do {                                                                            \
        __asm__ __volatile__("csrr %0, " #a : "=r"(a));                             \
        if (!strcmp(#a, "vtype")) {                                                 \
            printf("vtype(1.0) vill=%lu,vsew=%lu,vlmul=%lu,vta=%lu,vma=%lu\n",      \
                   1 & (a >> 63), 7 & (a >> 2), 3 & a, 1 & (a >> 5), 1 & (a >> 6)); \
            printf("vtype(0.9) vill=%lu,vsew=%lu,vlmul=%lu,vta=%lu,vma=%lu\n",      \
                   1 & (a >> 63), 7 & (a >> 2), 3 & a, 1 & (a >> 5), 1 & (a >> 6)); \
        } else                                                                      \
            printf("%s = 0x%lx\n", #a, a);                                          \
    } while (0);

    size_t vtype, vl, vlenb, vstart, vxrm, vxsat, vcsr;
    CSR(vtype);
    CSR(vl);
    CSR(vlenb);
    CSR(vstart);
    CSR(vxrm);
    CSR(vxsat);
}

 float vmaxabs(size_t sz, const float* t, float init) {
#ifdef RVV_JDSK
    asm volatile(
      "vsetvli        t0, x0, e32, m2     \n\t"
      "vfmv.v.f       v0, %[INIT]         \n\t"

      "MAIN_VMAXABS:                      \n\t"
      "vsetvli        t0, %[SZ], e32, m2  \n\t"
      "slli           t1, t0, 2           \n\t"
      "vle32.v        v2, (%[T])          \n\t"
      "vfsgnjx.vv     v4, v2, v2          \n\t"
      "vfmax.vv       v0, v4, v0          \n\t"
      "sub            %[SZ], %[SZ], t0    \n\t"
      "add            %[T], %[T], t1      \n\t"
      "bgtz           %[SZ], MAIN_VMAXABS \n\t"

      "vsetvli        t0, zero, e32, m1   \n\t"
      "vfmax.vv       v2, v0, v1          \n\t"
      "vfredmax.vs    v4, v2, v0          \n\t"
      "vfmv.f.s       %[INIT], v4         \n\t"
      :[T]"+r"(t), [INIT]"+f"(init)
      :[SZ]"r"(sz)
      :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5", "v6");
#else
    VSET1(e32, m1);
    asm volatile("vfmv.s.f v1, %[init]\n" : : [init] "f"(init));
    size_t lmul = mk_lmul(E32, sz);
    size_t vt8 = mk_vtype(E8, lmul), vt32 = mk_vtype(E32, lmul);
    for (; sz > 0;) {
        int vl;
        asm volatile(
                "vsetvl        x0, %[sz8], %[vt8]\n"
                "vlbu.v        v8, (%[t])\n"
                "vsetvl        %[vl], %[sz32], %[vt32]\n"
                "vfsgnjx.vv    v8, v8, v8\n"
                "vfredmax.vs   v1, v8, v1\n"
                : [vl] "=r"(vl)
                : [sz8] "r"(sz * 4), [vt8] "r"(vt8), [sz32] "r"(sz), [vt32] "r"(vt32),
                  [t] "r"(t)
                : "memory");
        t += vl;
        sz -= vl;
    }
    VSET1(e32, m1);
    asm volatile("vfmv.f.s  %[init],  v1\n" : [init] "=f"(init));
#endif
    return init;
}

 float vmax(int sz, const float* t, float init) {
#ifdef RVV_JDSK
    asm volatile(
      "vsetvli        t0, x0, e32, m2     \n\t"
      "vfmv.v.f       v0, %[INIT]         \n\t"

      "MAIN_VMAX:                         \n\t"
      "vsetvli        t0, %[SZ], e32, m2  \n\t"
      "slli           t1, t0, 2           \n\t"
      "vle32.v        v2, (%[T])          \n\t"
      "vfmax.vv       v0, v4, v0          \n\t"
      "sub            %[SZ], %[SZ], t0    \n\t"
      "add            %[T], %[T], t1      \n\t"
      "bgtz           %[SZ], MAIN_VMAX    \n\t"

      "vsetvli        t0, zero, e32, m1   \n\t"
      "vfmax.vv       v2, v0, v1          \n\t"
      "vfredmax.vs    v4, v2, v0          \n\t"
      "vfmv.f.s       %[INIT], v4         \n\t"
      :[T]"+r"(t), [INIT]"+f"(init)
      :[SZ]"r"(sz)
      :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5", "v6");
#else
    VSET1(e32, m1);
    asm volatile("vfmv.s.f v1, %[init]\n" : : [init] "f"(init));
    size_t lmul = mk_lmul(E32, sz);
    size_t vt8 = mk_vtype(E8, lmul), vt32 = mk_vtype(E32, lmul);
    for (; sz > 0;) {
        int vl;
        asm volatile(
                "vsetvl        x0, %[sz8], %[vt8]\n"
                "vlbu.v        v8, (%[t])\n"
                "vsetvl        %[vl], %[sz32], %[vt32]\n"
                "vfredmax.vs   v1, v8, v1\n"
                : [vl] "=r"(vl)
                : [sz8] "r"(sz * 4), [vt8] "r"(vt8), [sz32] "r"(sz), [vt32] "r"(vt32),
                  [t] "r"(t)
                : "memory");
        t += vl;
        sz -= vl;
    }
    VSET1(e32, m1);
    asm volatile("vfmv.f.s  %[init],  v1\n" : [init] "=f"(init));
#endif
    return init;
}

 float vmulsum(const float* x, const float* y, int sz, float init) {
#ifdef RVV_JDSK
    asm volatile(
        "vsetvli        t0, x0, e32, m2     \n\t"
        "vmv.v.i        v8, 0               \n\t"
        "vfmv.s.f       v0, %[INIT]         \n\t"
        "MAIN_VMULSUM:                      \n\t"

        "vsetvli        t0, %[SZ], e32, m2  \n\t"
        "slli           t1, t0, 2           \n\t"
        "vle32.v        v2, (%[X])          \n\t"
        "vle32.v        v4, (%[Y])          \n\t"
        "vfmacc.vv      v8, v2, v4          \n\t"
        "sub            %[SZ], %[SZ], t0    \n\t"
        "add            %[X], %[X], t1      \n\t"
        "add            %[Y], %[Y], t1      \n\t"
        "bgtz           %[SZ], MAIN_VMULSUM \n\t"

        "vsetvli        t0, zero, e32, m1   \n\t"
        "vfadd.vv       v2, v8, v9          \n\t"
        "vfredsum.vs    v4, v2, v0          \n\t"
        "vfmv.f.s       %[INIT], v4         \n\t"
        :[X]"+r"(x), [Y]"+r"(y), [INIT]"+f"(init)
        :[SZ]"r"(sz)
        :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5",
          "v6", "v7", "v8", "v9");
#else
    asm volatile(
            "vsetvli x0, %[sz], e32, m1\n"
            "vfmv.s.f v1, %[init]\n"
            :
            : [init] "f"(init), [sz] "r"(sz));
    for (; sz > 0;) {
        int vl = 0;
        asm volatile(
                "slli          t0, %[sz], 2\n"
                "vsetvli       t0, t0, e8, m8\n"
                "vlbu.v        v8,  (%[x])\n"
                "vlbu.v        v16, (%[y])\n"
                "srli          t0, t0, 2\n"
                "vsetvli       %[vl], t0, e32, m8\n"
                "vfmul.vv      v8, v8, v16\n"
                "vfredsum.vs   v1, v8, v1\n"
                : [vl] "=r"(vl)
                : [sz] "r"(sz), [x] "r"(x), [y] "r"(y)
                : "t0", "memory");
        x += vl;
        y += vl;
        sz -= vl;
    }
    asm volatile(
            "vsetvli x0, x0, e32, m1\n"
            "vfmv.f.s  %[init],  v1\n"
            : [init] "=f"(init));
#endif
    return init;
}

 float vsqrsum(const float* t, int sz, float init) {
#ifdef RVV_JDSK
    asm volatile(
        "vsetvli        t0, x0, e32, m2     \n\t"
        "vmv.v.i        v6, 0               \n\t"
        "vfmv.s.f       v0, %[INIT]         \n\t"
        "MAIN_VSQRSUM:                      \n\t"

        "vsetvli        t0, %[SZ], e32, m2  \n\t"
        "slli           t1, t0, 2           \n\t"
        "vle32.v        v2, (%[T])          \n\t"
        "vfmacc.vv      v6, v2, v2          \n\t"
        "sub            %[SZ], %[SZ], t0    \n\t"
        "add            %[T], %[T], t1      \n\t"
        "bgtz           %[SZ], MAIN_VSQRSUM \n\t"

        "vsetvli        t0, zero, e32, m1   \n\t"
        "vfadd.vv       v2, v6, v7          \n\t"
        "vfredsum.vs    v4, v2, v0          \n\t"
        "vfmv.f.s       %[INIT], v4         \n\t"
        :[T]"+r"(t), [INIT]"+f"(init)
        :[SZ]"r"(sz)
        :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5",
          "v6", "v7");
#else
    asm volatile(
            "vsetvli x0, %[sz], e32, m1\n"
            "vfmv.s.f v1, %[init]\n"
            :
            : [init] "f"(init), [sz] "r"(sz));
    for (; sz > 0;) {
        int vl = 0;
        asm volatile(
                "slli          t0, %[sz], 2\n"
                "vsetvli       t0, t0, e8, m8\n"
                "vlbu.v        v8, (%[t])\n"
                "srli          t0, t0, 2\n"
                "vsetvli       %[vl], t0, e32, m8\n"
                "vfmul.vv      v8, v8, v8\n"
                "vfredsum.vs   v1, v8, v1\n"
                : [vl] "=r"(vl)
                : [sz] "r"(sz), [t] "r"(t)
                : "t0", "memory");
        t += vl;
        sz -= vl;
    }
    asm volatile(
            "vsetvli x0, x0, e32, m1\n"
            "vfmv.f.s  %[init],  v1\n"
            : [init] "=f"(init));
#endif
    return init;
}

 void vscal(int sz, const float* x, float* z, float scale) {
#ifdef RVV_JDSK
    asm volatile(
            "MAIN_VSCAL:                        \n\t"

            "vsetvli        t0, %[SZ], e32, m2  \n\t"
            "slli           t1, t0, 2           \n\t"
            "vle32.v        v0, (%[X])          \n\t"
            "vfmul.vf       v2, v0, %[SCALE]    \n\t"
            "sub            %[SZ], %[SZ], t0    \n\t"
            "add            %[X], %[X], t1      \n\t"
            "vse32.v        v2, (%[Z])          \n\t"
            "add            %[Z], %[Z], t1      \n\t"
            "bgtz           %[SZ], MAIN_VSCAL   \n\t"

            :[X]"+r"(x), [Z]"+r"(z)
            :[SCALE]"f"(scale), [SZ]"r"(sz)
            :"cc", "t0", "t1", "v0", "v1", "v2", "v3");
#else

    for (; sz > 0;) {
        int vl = 0;
        asm volatile(
                "slli          t0, %[sz], 2\n"
                "vsetvli       t0, t0, e8, m8\n"
                "vlbu.v        v8, (%[x])\n"
                "srli          t1, t0, 2\n"
                "vsetvli       %[vl], t1, e32, m8\n"
                "vfmul.vf      v8, v8, %[scale]\n"
                "vsetvli       x0, t0, e8, m8\n"
                "vsb.v         v8, (%[z])\n"
                : [vl] "=r"(vl)
                : [sz] "r"(sz), [x] "r"(x), [z] "r"(z), [scale] "f"(scale)
                : "t0", "t1", "memory");
        x += vl;
        z += vl;
        sz -= vl;
    }
#endif
}

 void vadd(int sz, const float* x, const float* y, float* z) {
#ifdef RVV_JDSK
    asm volatile(
            "MAIN_VADD:                         \n\t"

            "vsetvli        t0, %[SZ], e32, m2  \n\t"
            "slli           t1, t0, 2           \n\t"
            "vle32.v        v0, (%[X])          \n\t"
            "vle32.v        v2, (%[Y])          \n\t"
            "vfadd.vv       v4, v0, v2          \n\t"
            "sub            %[SZ], %[SZ], t0    \n\t"
            "add            %[X], %[X], t1      \n\t"
            "add            %[Y], %[Y], t1      \n\t"
            "vse32.v        v4, (%[Z])          \n\t"
            "add            %[Z], %[Z], t1      \n\t"
            "bgtz           %[SZ], MAIN_VADD    \n\t"

            :[X]"+r"(x), [Y]"+r"(y), [Z]"+r"(z)
            :[SZ]"r"(sz)
            :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5");
#else
    for (; sz > 0;) {
        int vl = 0;
        asm volatile(
                "slli          t0, %[sz], 2\n"
                "vsetvli       t0, t0, e8, m8\n"
                "vlbu.v        v8, (%[x])\n"
                "vlbu.v        v16, (%[y])\n"
                "srli          t1, t0, 2\n"
                "vsetvli       %[vl], t1, e32, m8\n"
                "vfadd.vv      v8, v8, v16\n"
                "vsetvli       x0, t0, e8, m8\n"
                "vsb.v         v8, (%[z])\n"
                : [vl] "=r"(vl)
                : [sz] "r"(sz), [x] "r"(x), [y] "r"(y), [z] "r"(z)
                : "t0", "t1", "memory");
        x += vl;
        y += vl;
        z += vl;
        sz -= vl;
    }
#endif
}

 void vmul(int sz, const float* x, const float* y, float* z) {
#ifdef RVV_JDSK
    asm volatile(
            "MAIN_VMUL:                         \n\t"

            "vsetvli        t0, %[SZ], e32, m2  \n\t"
            "slli           t1, t0, 2           \n\t"
            "vle32.v        v0, (%[X])          \n\t"
            "vle32.v        v2, (%[Y])          \n\t"
            "vfmul.vv       v4, v0, v2          \n\t"
            "sub            %[SZ], %[SZ], t0    \n\t"
            "add            %[X], %[X], t1      \n\t"
            "add            %[Y], %[Y], t1      \n\t"
            "vse32.v        v4, (%[Z])          \n\t"
            "add            %[Z], %[Z], t1      \n\t"
            "bgtz           %[SZ], MAIN_VMUL    \n\t"

            :[X]"+r"(x), [Y]"+r"(y), [Z]"+r"(z)
            :[SZ]"r"(sz)
            :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5");
#else

    for (; sz > 0;) {
        int vl = 0;
        asm volatile(
                "slli          t0, %[sz], 2\n"
                "vsetvli       t0, t0, e8, m8\n"
                "vlbu.v        v8, (%[x])\n"
                "vlbu.v        v16, (%[y])\n"
                "srli          t1, t0, 2\n"
                "vsetvli       %[vl], t1, e32, m8\n"
                "vfmul.vv      v8, v8, v16\n"
                "vsetvli       x0, t0, e8, m8\n"
                "vsb.v         v8, (%[z])\n"
                : [vl] "=r"(vl)
                : [sz] "r"(sz), [x] "r"(x), [y] "r"(y), [z] "r"(z)
                : "t0", "t1", "memory");
        x += vl;
        y += vl;
        z += vl;
        sz -= vl;
    }
#endif
}

}  // namespace opt
}  // namespace inferllm
