#pragma once

#include <assert.h>
#include <cstdlib>
#include "core/tensor.h"
#include "file.h"
#include "kern/kernel_define.h"

namespace inferllm {
namespace opt {

typedef enum SEW {
    E8 = 0,
    E16 = 1,
    E32 = 2,
    E64 = 3,
} SEW;

extern size_t vlmax[4][4];
extern size_t vlmul[4][4];

inline size_t mk_lmul(SEW sew, int sz) {
    size_t rvlmul;
    for (int i = 0; i < 4; i++) {
        rvlmul = vlmul[sew][i];
        if (sz <= vlmax[sew][i])
            break;
    }
    return rvlmul;
}

inline size_t mk_vtype(SEW sew, size_t LMUL) {
#if INFER_RVV > 107
    return (sew << 3) | LMUL;
#else
    return (sew << 2) | LMUL;
#endif
}
#ifndef RVV_JDSK
#if INFER_RVV > 107
#define VSET1(e, m) asm volatile("vsetivli x0, 1, " #e ", " #m);
#else
#define VSET1(e, m) asm volatile("vsetvli x0, %[sz], " #e ", " #m ::[sz] "r"(1));
#endif
#endif

float vmaxabs(size_t sz, const float* t, float init);
float vmax(int sz, const float* t, float init);
float vmulsum(const float* x, const float* y, int sz, float init);
float vsqrsum(const float* t, int sz, float init);
void vscal(int sz, const float* x, float* z, float scale);
void vadd(int sz, const float* x, const float* y, float* z);
void vmul(int sz, const float* x, const float* y, float* z);

void dumpV();

}  // namespace opt
}  // namespace inferllm
