#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <assert.h>

float vmaxabs(size_t sz, const float* t, float init) {
    asm volatile(
      "vsetvli        t0, x0, e32, m2     \n\t"
      "vfmv.v.f       v0, %[INIT]         \n\t"

      "MAIN_VMAXABS:                      \n\t"
      "vsetvli        t0, %[SZ], e32, m2  \n\t"
      "slli           t1, t0, 2           \n\t"
      "vle32.v        v2, (%[T])          \n\t"
      "vfsgnjx.vv     v4, v2, v2          \n\t"
      "vfmax.vv       v0, v4, v0          \n\t"
      "sub            %[SZ], %[SZ], t0    \n\t"
      "add            %[T], %[T], t1      \n\t"
      "bgtz           %[SZ], MAIN_VMAXABS \n\t"

      "vsetvli        t0, zero, e32, m1   \n\t"
      "vfmax.vv       v2, v0, v1          \n\t"
      "vfredmax.vs    v4, v2, v0          \n\t"
      "vfmv.f.s       %[INIT], v4         \n\t"
      :[T]"+r"(t), [INIT]"+f"(init)
      :[SZ]"r"(sz)
      :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5", "v6");
    return init;
}

float vmaxabs_ref(size_t sz, const float* t, float init) {
    for (int i = 0; i  < sz; i++)
    {
        float tmp = fabs(*t++);
        init = init > tmp ? init : tmp;
    }

    return init;
}

float vmax(int sz, const float* t, float init) {
    asm volatile(
      "vsetvli        t0, x0, e32, m2     \n\t"
      "vfmv.v.f       v0, %[INIT]         \n\t"

      "MAIN_VMAX:                         \n\t"
      "vsetvli        t0, %[SZ], e32, m2  \n\t"
      "slli           t1, t0, 2           \n\t"
      "vle32.v        v2, (%[T])          \n\t"
      "vfmax.vv       v0, v4, v0          \n\t"
      "sub            %[SZ], %[SZ], t0    \n\t"
      "add            %[T], %[T], t1      \n\t"
      "bgtz           %[SZ], MAIN_VMAX    \n\t"

      "vsetvli        t0, zero, e32, m1   \n\t"
      "vfmax.vv       v2, v0, v1          \n\t"
      "vfredmax.vs    v4, v2, v0          \n\t"
      "vfmv.f.s       %[INIT], v4         \n\t"
      :[T]"+r"(t), [INIT]"+f"(init)
      :[SZ]"r"(sz)
      :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5", "v6");
    return init;
}

float vmax_ref(size_t sz, const float* t, float init) {
    for (int i = 0; i  < sz; i++)
    {
        float tmp = *t++;
        init = init > tmp ? init : tmp;
    }

    return init;
}

float vmulsum(const float* x, const float* y, int sz, float init) {
    asm volatile(
        "vsetvli        t0, x0, e32, m2     \n\t"
        "vmv.v.i        v8, 0               \n\t"
        "vfmv.s.f       v0, %[INIT]         \n\t"
        "MAIN_VMULSUM:                      \n\t"

        "vsetvli        t0, %[SZ], e32, m2  \n\t"
        "slli           t1, t0, 2           \n\t"
        "vle32.v        v2, (%[X])          \n\t"
        "vle32.v        v4, (%[Y])          \n\t"
        "vfmacc.vv      v8, v2, v4          \n\t"
        "sub            %[SZ], %[SZ], t0    \n\t"
        "add            %[X], %[X], t1      \n\t"
        "add            %[Y], %[Y], t1      \n\t"
        "bgtz           %[SZ], MAIN_VMULSUM \n\t"

        "vsetvli        t0, zero, e32, m1   \n\t"
        "vfadd.vv       v2, v8, v9          \n\t"
        "vfredsum.vs    v4, v2, v0          \n\t"
        "vfmv.f.s       %[INIT], v4         \n\t"
        :[X]"+r"(x), [Y]"+r"(y), [INIT]"+f"(init)
        :[SZ]"r"(sz)
        :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5",
          "v6", "v7", "v8", "v9");
    return init;
}

float vmulsum_ref(const float* x, const float* y, int sz, float init) {
    for (int i = 0; i < sz; i++)
    {
        float tmp = (*x++)*(*y++);
        init += tmp;
    }
    return init;
}

float vsqrsum(const float* t, int sz, float init) {
    asm volatile(
        "vsetvli        t0, x0, e32, m2     \n\t"
        "vmv.v.i        v6, 0               \n\t"
        "vfmv.s.f       v0, %[INIT]         \n\t"
        "MAIN_VSQRSUM:                      \n\t"

        "vsetvli        t0, %[SZ], e32, m2  \n\t"
        "slli           t1, t0, 2           \n\t"
        "vle32.v        v2, (%[T])          \n\t"
        "vfmacc.vv      v6, v2, v2          \n\t"
        "sub            %[SZ], %[SZ], t0    \n\t"
        "add            %[T], %[T], t1      \n\t"
        "bgtz           %[SZ], MAIN_VSQRSUM \n\t"

        "vsetvli        t0, zero, e32, m1   \n\t"
        "vfadd.vv       v2, v6, v7          \n\t"
        "vfredsum.vs    v4, v2, v0          \n\t"
        "vfmv.f.s       %[INIT], v4         \n\t"
        :[T]"+r"(t), [INIT]"+f"(init)
        :[SZ]"r"(sz)
        :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5",
          "v6", "v7");
    return init;
}

float vsqrsum_ref(const float* t, int sz, float init) {
    for (int i = 0; i < sz; i++)
    {
        float tmp = (*t)*(*t);
        t++;
        init += tmp;
    }
    return init;
}


void vscal(int sz, const float* x, float* z, float scale) {
    asm volatile(
        "MAIN_VSCAL:                        \n\t"

        "vsetvli        t0, %[SZ], e32, m2  \n\t"
        "slli           t1, t0, 2           \n\t"
        "vle32.v        v0, (%[X])          \n\t"
        "vfmul.vf       v2, v0, %[SCALE]    \n\t"
        "sub            %[SZ], %[SZ], t0    \n\t"
        "add            %[X], %[X], t1      \n\t"
        "vse32.v        v2, (%[Z])          \n\t"
        "add            %[Z], %[Z], t1      \n\t"
        "bgtz           %[SZ], MAIN_VSCAL   \n\t"

        :[X]"+r"(x), [Z]"+r"(z)
        :[SCALE]"f"(scale), [SZ]"r"(sz)
        :"cc", "t0", "t1", "v0", "v1", "v2", "v3");
}

void vscal_ref(int sz, const float* x, float* z, float scale) {
    for (int i = 0; i < sz; i++){
        *z++ = (*x++) * scale;
    }
}

void vadd(int sz, const float* x, const float* y, float* z) {
    asm volatile(
       "MAIN_VADD:                         \n\t"

       "vsetvli        t0, %[SZ], e32, m2  \n\t"
       "slli           t1, t0, 2           \n\t"
       "vle32.v        v0, (%[X])          \n\t"
       "vle32.v        v2, (%[Y])          \n\t"
       "vfadd.vv       v4, v0, v2          \n\t"
       "sub            %[SZ], %[SZ], t0    \n\t"
       "add            %[X], %[X], t1      \n\t"
       "add            %[Y], %[Y], t1      \n\t"
       "vse32.v        v4, (%[Z])          \n\t"
       "add            %[Z], %[Z], t1      \n\t"
       "bgtz           %[SZ], MAIN_VADD    \n\t"

       :[X]"+r"(x), [Y]"+r"(y), [Z]"+r"(z)
       :[SZ]"r"(sz)
       :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5");
}

void vadd_ref(int sz, const float* x, float* y, float* z) {
    for (int i = 0; i < sz; i++){
        *z++ = (*x++) + (*y++);
    }
}


void vmul(int sz, const float* x, const float* y, float* z) {
    asm volatile(
        "MAIN_VMUL:                         \n\t"

        "vsetvli        t0, %[SZ], e32, m2  \n\t"
        "slli           t1, t0, 2           \n\t"
        "vle32.v        v0, (%[X])          \n\t"
        "vle32.v        v2, (%[Y])          \n\t"
        "vfmul.vv       v4, v0, v2          \n\t"
        "sub            %[SZ], %[SZ], t0    \n\t"
        "add            %[X], %[X], t1      \n\t"
        "add            %[Y], %[Y], t1      \n\t"
        "vse32.v        v4, (%[Z])          \n\t"
        "add            %[Z], %[Z], t1      \n\t"
        "bgtz           %[SZ], MAIN_VMUL    \n\t"

        :[X]"+r"(x), [Y]"+r"(y), [Z]"+r"(z)
        :[SZ]"r"(sz)
        :"cc", "t0", "t1", "v0", "v1", "v2", "v3", "v4", "v5");
}

void vmul_ref(int sz, const float* x, float* y, float* z) {
    for (int i = 0; i < sz; i++){
        *z++ = (*x++) * (*y++);
    }
}

#define QK40 32
struct BlockQ40 {
    float d;               // delta
    uint8_t qs[QK40 / 2];  // nibbles / quants
}BlockQ40_t;
//static_assert(sizeof(BlockQ40) == 20, "BlockQ40 size error");

struct BlockQ40X8 {
    uint8_t qs[QK40 / 2 * 8];     // nibbles / quants
    float scale[8];               // delta
};
//static_assert(sizeof(BlockQ40X8) == 160, "BlockQ40X8 size error");

#define QK80 32
struct BlockQ80 {
    float d;          // delta
    int8_t qs[QK80];  // nibbles
}BlockQ80_t;

float vec_vec_dot_q40_with_q80_ref(
        const int n, const void* vx, const void* vy) {
    const int nb = n / QK80;
    assert(n % QK80 == 0);
    assert(nb % 2 == 0);

    const struct BlockQ40*  x = (struct BlockQ40*)(vx);
    const struct BlockQ80*  y = (struct BlockQ80*)(vy);

    float sumf = 0.0;
    for (int i = 0; i < nb; i++) {
        const float d0 = x[i].d;
        const float d1 = y[i].d;

        const uint8_t*  p0 = x[i].qs;
        const int8_t*  p1 = y[i].qs;

        int sumi = 0;
        for (int j = 0; j < QK80 / 2; j++) {
            const uint8_t v0 = p0[j];

            const int i0 = (int8_t)(v0 & 0x0F) - 8;
            const int i1 = (int8_t)(v0 >> 4) - 8;

            const int i2 = p1[2 * j + 0];
            const int i3 = p1[2 * j + 1];

            sumi += i0 * i2 + i1 * i3;
        }
        sumf += d0 * d1 * sumi;
    }
    return sumf;
}

float vec_vec_dot_q40_with_q80(
        const int n, const void* vx, const void* vy) {
    const int nb = n / QK80;
    assert(n % QK80 == 0);
    assert(nb % 2 == 0);

    const struct BlockQ40*  x = (struct BlockQ40*)(vx);
    const struct BlockQ80*  y = (struct BlockQ80*)(vy);

    float sumf = 0.0;
    for (int i = 0; i < nb; i++) {
        const float d0 = x[i].d;
        const float d1 = y[i].d;

        const uint8_t*  p0 = x[i].qs;
        const int8_t*  p1 = y[i].qs;

        int sumi = 0;
        asm volatile(
           "li             t1, 8              \n\t"
           "vsetvli        t0, zero, e8, mf2  \n\t"
           "vle8.v         v0, (%[X])         \n\t"
           "vwaddu.vx      v2, v0, zero       \n\t"
           "vsetvli        t0, zero, e16, m1  \n\t"
           "vwaddu.vx      v4, v2, zero       \n\t"
           "vsetvli        t0, zero, e32, m2  \n\t"
           "vand.vi        v8, v4, 0xf        \n\t"   // low  4bit
           "vsrl.vi        v0, v4, 4          \n\t"   // high 4bit
           "vsll.vi        v12, v0, 16        \n\t"
           "vor.vv         v8, v8, v12        \n\t"

           "vmv.v.i        v20, 0             \n\t"
           "vsetvli        t0, zero, e8, m1   \n\t"
           "vle8.v         v0, (%[Y])         \n\t"
           "vwadd.vx       v4, v0, zero       \n\t"
           "vsetvli        t0, zero, e16, m2  \n\t"
           "vsub.vx        v12, v8, t1        \n\t"
           "vwmul.vv       v16, v12, v4       \n\t"
           "vsetvli        t0, zero, e32, m2  \n\t"
           "vadd.vv        v8, v16, v18       \n\t"
           "vsetvli        t0, zero, e32, m1  \n\t"
           "vadd.vv        v4, v8, v9         \n\t"
           "vredsum.vs     v6, v4, v20        \n\t"
           "vmv.x.s        %[SUM], v6         \n\t"
           :[X]"+r"(p0), [Y]"+r"(p1), [SUM]"+r"(sumi)
           :
           :"cc", "t0", "t1");
        sumf += d0 * d1 * sumi;
    }
    return sumf;
}

#define LENGTH 1023
#define NB 12

float src0[LENGTH] = {};
float src1[LENGTH] = {};
float ref [LENGTH] = {};
float dst [LENGTH] = {};

int main()
{
    srand(time(NULL));

    for (int i = 0; i < LENGTH; i++)
    {
        src0[i] = rand()% 511 + 5;
        src1[i] = rand()% 127 + 5;
    }

    float abs_max = vmaxabs(LENGTH, src0, 0.0);
    float abs_max_ref = vmaxabs_ref(LENGTH, src0, 0.0);

    float max = vmax(LENGTH, src0, 0.0);
    float max_ref = vmax_ref(LENGTH, src0, 0.0);

    float vmul_sum = vmulsum(src0, src1, LENGTH, 0.0);
    float vmul_sum_ref = vmulsum_ref(src0, src1, LENGTH, 0.0);

    float vsqr_sum = vsqrsum(src0, LENGTH, 0.0);
    float vsqr_sum_ref = vsqrsum_ref(src0, LENGTH, 0.0);

    vscal_ref(LENGTH, src0, ref, 2.0);
    vscal(LENGTH, src0, dst, 2.0);

    for (int i = 0; i < LENGTH; i++)
    {
        if (ref[i] != dst[i])
        {
            printf ("vscale error, the ref is %lf, the result is %lf.\n", ref[i], dst[i]);
        }
    }

    vadd_ref(LENGTH, src0, src1, ref);
    vadd(LENGTH, src0, src1, dst);

    for (int i = 0; i < LENGTH; i++)
    {
        if (ref[i] != dst[i])
        {
            printf ("vadd error, the ref is %lf, the result is %lf.\n", ref[i], dst[i]);
        }
    }

    vmul_ref(LENGTH, src0, src1, ref);
    vmul(LENGTH, src0, src1, dst);

    for (int i = 0; i < LENGTH; i++)
    {
        if (ref[i] != dst[i])
        {
            printf ("vmul error, the ref is %lf, the result is %lf.\n", ref[i], dst[i]);
        }
    }

    if (abs_max != abs_max_ref)
    {
        printf ("abs max error.\n");
    }

    if (max != max_ref)
    {
        printf ("max error.\n");
    }

    if (vmul_sum != vmul_sum_ref)
    {
        printf ("vmulsum error, the ref is %lf, the result is %lf.\n", vmul_sum_ref, vmul_sum);
    }

    if (vsqr_sum != vsqr_sum_ref)
    {
        printf ("vsqrsum error, the ref is %lf, the result is %lf.\n", vsqr_sum_ref, vsqr_sum);
    }


    struct BlockQ40 bq4[NB];
    struct BlockQ80 bq8[NB];

    for (int i = 0; i < NB; i++)
    {
        bq4[i].d = rand()% 63 + 5;
        bq8[i].d = rand()% 63 + 5;
        for (int j = 0; j < 32; j++)
        {
            bq4[i].qs[j] = rand()% 511 + 5;
            bq8[i].qs[j] = rand()% 127 + 5;
        }
    }

    float dot_ref = vec_vec_dot_q40_with_q80_ref(NB*32, bq4, bq8);
    float dot = vec_vec_dot_q40_with_q80(NB*32, bq4, bq8);
    
    if (dot_ref != dot)
    {
        printf ("vec dot error, the ref is %lf, the result is %lf.\n", dot_ref, dot);
    }

    return 0;
}
